#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
import dataclasses
from typing import List, Tuple, Any, Dict
import enum
from dataclasses import dataclass
from datetime import datetime, timedelta, date, time, timezone
import json


class MaintenanceState(enum.Enum):
    Active = enum.auto()
    PreMaint = enum.auto()
    Maint = enum.auto()
    Unknown = enum.auto()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s: str):
        for ms in MaintenanceState:
            if s == ms.name:
                return cls(ms)
        return None


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if isinstance(o, MaintenanceState):
            return str(o)
        if isinstance(o, (datetime, date, time)):
            return o.isoformat()
        return super().default(o)


@dataclass(eq=True)
class MaintenanceEntry:
    state: MaintenanceState
    deadline: datetime = None # when it starts
    expected_end: datetime = None # how long it may be

    def __init__(self, state: MaintenanceState or str,
                 deadline: datetime or str or None = None,
                 expected_end: datetime or str or None = None):
        if isinstance(state, MaintenanceState):
            self.state = state
        else:
            self.state = MaintenanceState.from_string(state)

        if isinstance(deadline, datetime):
            self.deadline = deadline
        elif deadline:
            self.deadline = datetime.fromisoformat(deadline)

        if isinstance(expected_end, datetime):
            self.expected_end = expected_end
        elif expected_end:
            self.expected_end = datetime.fromisoformat(expected_end)


class MaintenanceModeException(Exception):
    def __init(self, msg: str):
        super().__init__(msg)


# ABQM maintenance information is a dictionary
class MaintenanceInfo:

    def __init__(self):
        self._nodes = dict()
        # prevent modifications
        self._lock = False

    def _set(self, d: Dict):
        self._nodes = d

    def finalize(self):
        """
        Prevent further modifications - used after the object is
        assigned to a property
        """
        self._lock = True

    def add(self, name: str, minfo: MaintenanceEntry) -> None:
        """
        Add entry
        """
        if self._lock:
            raise MaintenanceModeException("Unable to modify a finalized object, recreate and reassign")
        self._nodes[name] = minfo

    def get(self, name: str) -> MaintenanceEntry or None:
        return self._nodes.get(name)

    def rem(self, name: str) -> None:
        """
        Remove entry
        """
        if self._lock:
            raise MaintenanceModeException("Unable to modify a finalized object, recreate and reassign")
        self._nodes.pop(name)

    def pop(self, name: str) -> MaintenanceEntry:
        """
        Pop an entry returning maintenance info
        """
        if self._lock:
            raise MaintenanceModeException("Unable to modify a finalized object, recreate and reassign")
        return self._nodes.pop(name)

    def to_json(self) -> str:
        """
        Convert structure to JSON
        """
        if not self._lock:
            raise MaintenanceModeException("Object should be finalized prior to serializing")
        return json.dumps(self._nodes, cls=EnhancedJSONEncoder)

    def copy(self):
        """
        Copy an instance of the object but don't finalize
        """
        t = MaintenanceInfo()
        t._nodes = self._nodes.copy()
        return t

    def list_names(self) -> List[str]:
        """
        List the names of the nodes in maintenance
        """
        return list(self._nodes.keys())

    def list_details(self) -> List[Tuple[str, Any]]:
        """
        Return a list of tuples with node name and maintenance state details
        """
        return list(self._nodes.copy().items())

    def iter(self):
        """
        Generate an item iterator on a finalized object, so you can do
        for i in minfo.iter():
          print(i)
        """
        if not self._lock:
            raise MaintenanceModeException("Object should be finalized prior to attempting iteration")
        for i in self._nodes.items():
            yield i

    @classmethod
    def from_json(cls, json_string: str):
        """
        Convert into structure from JSON
        """
        if json_string is None or len(json_string) == 0:
            return None
        o = json.loads(json_string)
        ret = cls()
        ret._set({k: MaintenanceEntry(**v) for (k, v) in o.items()})
        ret.finalize()
        return ret

    def __str__(self):
        return self.to_json()




"""Utilities with networkx."""

from collections.abc import Iterable
from typing import Any, Callable, List, Tuple

from networkx import DiGraph, Graph, all_simple_paths
from networkx.utils import pairwise

__all__ = [
    "get_first_item",
    "get_second_item",
    "get_two_first_items",
    "filter_by_degree",
    "get_attributs",
    "search_root_nodes",
    "search_leaf_nodes",
    "search_simple_path",
]


def get_first_item(t: Tuple) -> Any:
    """Returns first item of tuple."""
    return t[0]


def get_second_item(t: Tuple) -> Any:  # pragma: no cover
    """Returns second item of tuple."""
    return t[1]


def get_two_first_items(t: Tuple) -> Tuple:
    """Return (first, second) item of tuple as tuple."""
    return (t[0], t[1])


def filter_by_degree(degree: int) -> Callable[[Any], bool]:  # pragma: no cover
    """Generate a filter for specified degree."""

    def _filter(node: Any) -> bool:
        (id, d) = node
        return d == degree

    return _filter


def get_attributs(*names: str) -> Callable[[Tuple], List]:  # pragma: no cover
    """Returns attributs list for specified node or edge."""

    def _map(t: Tuple) -> List:
        (v, d) = t
        return [d[name] for name in names]

    return _map


def search_root_nodes(graph: DiGraph) -> Iterable[Any]:  # pragma: no cover
    """Search nodes with no 'in' edges.

    Arguments:
        graph (DiGraph): networkx digraph instance

    Returns:
        (Iterable[Any]): results as an iterable of node identifier.

    """

    return map(
        get_first_item,
        filter(filter_by_degree(0), graph.in_degree()),  # pyright: ignore [reportCallIssue, reportArgumentType]
    )


def search_leaf_nodes(graph: DiGraph) -> Iterable[Any]:  # pragma: no cover
    """Search nodes with no 'out' edges.

    Arguments:
        graph (DiGraph): networkx digraph instance

    Returns:
        (Iterable[Any]): results as an iterable of node identifier.

    """

    return map(
        get_first_item,
        filter(filter_by_degree(0), graph.out_degree()),  # pyright: ignore [reportCallIssue, reportArgumentType]
    )


def search_simple_path(graph: Graph, source: Any, target: Any, cutoff=None):  # pragma: no cover
    """Returns a list of edges."""
    return map(pairwise, all_simple_paths(graph, source=target, target=target, cutoff=cutoff))



import io as StringIO
import re
import string
from typing import Dict, Iterable, List, Match, Optional, TextIO, Tuple

from .metrics_core import Metric
from .samples import Sample
from .validation import (
    _is_valid_legacy_metric_name, _validate_labelname, _validate_metric_name,
)


def text_string_to_metric_families(text: str) -> Iterable[Metric]:
    """Parse Prometheus text format from a unicode string.

    See text_fd_to_metric_families.
    """
    yield from text_fd_to_metric_families(StringIO.StringIO(text))


ESCAPE_SEQUENCES = {
    '\\\\': '\\',
    '\\n': '\n',
    '\\"': '"',
}


def replace_escape_sequence(match: Match[str]) -> str:
    return ESCAPE_SEQUENCES[match.group(0)]


HELP_ESCAPING_RE = re.compile(r'\\[\\n]')
ESCAPING_RE = re.compile(r'\\[\\n"]')


def _replace_help_escaping(s: str) -> str:
    return HELP_ESCAPING_RE.sub(replace_escape_sequence, s)


def _replace_escaping(s: str) -> str:
    return ESCAPING_RE.sub(replace_escape_sequence, s)


def _is_character_escaped(s: str, charpos: int) -> bool:
    num_bslashes = 0
    while (charpos > num_bslashes
           and s[charpos - 1 - num_bslashes] == '\\'):
        num_bslashes += 1
    return num_bslashes % 2 == 1


def parse_labels(labels_string: str, openmetrics: bool = False) -> Dict[str, str]:
    labels: Dict[str, str] = {}

    # Copy original labels
    sub_labels = labels_string.strip()
    if openmetrics and sub_labels and sub_labels[0] == ',':
        raise ValueError("leading comma: " + labels_string)
    try:
        # Process one label at a time
        while sub_labels:
            # The label name is before the equal, or if there's no equal, that's the
            # metric name.
            
            term, sub_labels = _next_term(sub_labels, openmetrics)
            if not term:
                if openmetrics:
                    raise ValueError("empty term in line: " + labels_string)
                continue
            
            quoted_name = False
            operator_pos = _next_unquoted_char(term, '=')
            if operator_pos == -1:
                quoted_name = True
                label_name = "__name__"
            else:
                value_start = _next_unquoted_char(term, '=')
                label_name, quoted_name = _unquote_unescape(term[:value_start])
                term = term[value_start + 1:]
                
            if not quoted_name and not _is_valid_legacy_metric_name(label_name):
                raise ValueError("unquoted UTF-8 metric name")
                
            # Check for missing quotes 
            term = term.strip()
            if not term or term[0] != '"':
                raise ValueError

            # The first quote is guaranteed to be after the equal.
            # Find the last unescaped quote.
            i = 1
            while i < len(term):
                i = term.index('"', i)
                if not _is_character_escaped(term[:i], i):
                    break
                i += 1

            # The label value is between the first and last quote
            quote_end = i + 1
            if quote_end != len(term):
                raise ValueError("unexpected text after quote: " + labels_string)
            label_value, _ = _unquote_unescape(term[:quote_end])
            if label_name == '__name__':
                _validate_metric_name(label_name)
            else:
                _validate_labelname(label_name)
            if label_name in labels:
                raise ValueError("invalid line, duplicate label name: " + labels_string)
            labels[label_name] = label_value
        return labels
    except ValueError:
        raise ValueError("Invalid labels: " + labels_string)
    

def _next_term(text: str, openmetrics: bool) -> Tuple[str, str]:
    """Extract the next comma-separated label term from the text.
    
    Returns the stripped term and the stripped remainder of the string, 
    including the comma.
    
    Raises ValueError if the term is empty and we're in openmetrics mode.
    """
    
    # There may be a leading comma, which is fine here.
    if text[0] == ',':
        text = text[1:]
        if not text:
            return "", ""
        if text[0] == ',':
            raise ValueError("multiple commas")
    splitpos = _next_unquoted_char(text, ',}')
    if splitpos == -1:
        splitpos = len(text)
    term = text[:splitpos]
    if not term and openmetrics:
        raise ValueError("empty term:", term)
    
    sublabels = text[splitpos:]
    return term.strip(), sublabels.strip()


def _next_unquoted_char(text: str, chs: str, startidx: int = 0) -> int:
    """Return position of next unquoted character in tuple, or -1 if not found.
    
    It is always assumed that the first character being checked is not already
    inside quotes.
    """
    i = startidx
    in_quotes = False
    if chs is None:
        chs = string.whitespace
    while i < len(text):
        if text[i] == '"' and not _is_character_escaped(text, i):
            in_quotes = not in_quotes
        if not in_quotes:
            if text[i] in chs:
                return i
        i += 1
    return -1


def _last_unquoted_char(text: str, chs: str) -> int:
    """Return position of last unquoted character in list, or -1 if not found."""
    i = len(text) - 1
    in_quotes = False
    if chs is None:
        chs = string.whitespace
    while i > 0:
        if text[i] == '"' and not _is_character_escaped(text, i):
            in_quotes = not in_quotes
            
        if not in_quotes:
            if text[i] in chs:
                return i
        i -= 1
    return -1


def _split_quoted(text, separator, maxsplit=0):
    """Splits on split_ch similarly to strings.split, skipping separators if
    they are inside quotes.
    """

    tokens = ['']
    x = 0
    while x < len(text):
        split_pos = _next_unquoted_char(text, separator, x)
        if split_pos == -1:
            tokens[-1] = text[x:]
            x = len(text)
            continue
        if maxsplit > 0 and len(tokens) > maxsplit:
            tokens[-1] = text[x:]
            break
        tokens[-1] = text[x:split_pos]
        x = split_pos + 1
        tokens.append('')
    return tokens


def _unquote_unescape(text):
    """Returns the string, and true if it was quoted."""
    if not text:
        return text, False
    quoted = False
    text = text.strip()
    if text[0] == '"':
        if len(text) == 1 or text[-1] != '"':
            raise ValueError("missing close quote")
        text = text[1:-1]
        quoted = True
    if "\\" in text:
        text = _replace_escaping(text)
    return text, quoted


# If we have multiple values only consider the first
def _parse_value_and_timestamp(s: str) -> Tuple[float, Optional[float]]:
    s = s.lstrip()
    separator = " "
    if separator not in s:
        separator = "\t"
    values = [value.strip() for value in s.split(separator) if value.strip()]
    if not values:
        return float(s), None
    value = _parse_value(values[0])
    timestamp = (_parse_value(values[-1]) / 1000) if len(values) > 1 else None
    return value, timestamp


def _parse_value(value):
    value = ''.join(value)
    if value != value.strip() or '_' in value:
        raise ValueError(f"Invalid value: {value!r}")
    try:
        return int(value)
    except ValueError:
        return float(value)
    

def _parse_sample(text):
    separator = " # "
    # Detect the labels in the text
    label_start = _next_unquoted_char(text, '{')
    if label_start == -1 or separator in text[:label_start]:
        # We don't have labels, but there could be an exemplar.
        name_end = _next_unquoted_char(text, ' \t')
        name = text[:name_end].strip()
        if not _is_valid_legacy_metric_name(name):
            raise ValueError("invalid metric name:" + text)
        # Parse the remaining text after the name
        remaining_text = text[name_end + 1:]
        value, timestamp = _parse_value_and_timestamp(remaining_text)
        return Sample(name, {}, value, timestamp)
    name = text[:label_start].strip()
    label_end = _next_unquoted_char(text, '}')
    labels = parse_labels(text[label_start + 1:label_end], False)
    if not name:
        # Name might be in the labels
        if '__name__' not in labels:
            raise ValueError
        name = labels['__name__']
        del labels['__name__']
    elif '__name__' in labels:
        raise ValueError("metric name specified more than once")
    # Parsing labels succeeded, continue parsing the remaining text
    remaining_text = text[label_end + 1:]
    value, timestamp = _parse_value_and_timestamp(remaining_text)
    return Sample(name, labels, value, timestamp)


def text_fd_to_metric_families(fd: TextIO) -> Iterable[Metric]:
    """Parse Prometheus text format from a file descriptor.

    This is a laxer parser than the main Go parser,
    so successful parsing does not imply that the parsed
    text meets the specification.

    Yields Metric's.
    """
    name = ''
    documentation = ''
    typ = 'untyped'
    samples: List[Sample] = []
    allowed_names = []

    def build_metric(name: str, documentation: str, typ: str, samples: List[Sample]) -> Metric:
        # Munge counters into OpenMetrics representation
        # used internally.
        if typ == 'counter':
            if name.endswith('_total'):
                name = name[:-6]
            else:
                new_samples = []
                for s in samples:
                    new_samples.append(Sample(s[0] + '_total', *s[1:]))
                    samples = new_samples
        metric = Metric(name, documentation, typ)
        metric.samples = samples
        return metric

    for line in fd:
        line = line.strip()

        if line.startswith('#'):
            parts = _split_quoted(line, None, 3)
            if len(parts) < 2:
                continue
            candidate_name, quoted = '', False
            if len(parts) > 2:
                # Ignore comment tokens
                if parts[1] != 'TYPE' and parts[1] != 'HELP':
                    continue
                candidate_name, quoted = _unquote_unescape(parts[2])
                if not quoted and not _is_valid_legacy_metric_name(candidate_name):
                    raise ValueError
            if parts[1] == 'HELP':
                if candidate_name != name:
                    if name != '':
                        yield build_metric(name, documentation, typ, samples)
                    # New metric
                    name = candidate_name
                    typ = 'untyped'
                    samples = []
                    allowed_names = [candidate_name]
                if len(parts) == 4:
                    documentation = _replace_help_escaping(parts[3])
                else:
                    documentation = ''
            elif parts[1] == 'TYPE':
                if len(parts) < 4:
                    raise ValueError
                if candidate_name != name:
                    if name != '':
                        yield build_metric(name, documentation, typ, samples)
                    # New metric
                    name = candidate_name
                    documentation = ''
                    samples = []
                typ = parts[3]
                allowed_names = {
                    'counter': [''],
                    'gauge': [''],
                    'summary': ['_count', '_sum', ''],
                    'histogram': ['_count', '_sum', '_bucket'],
                }.get(typ, [''])
                allowed_names = [name + n for n in allowed_names]
        elif line == '':
            # Ignore blank lines
            pass
        else:
            sample = _parse_sample(line)
            if sample.name not in allowed_names:
                if name != '':
                    yield build_metric(name, documentation, typ, samples)
                # New metric, yield immediately as untyped singleton
                name = ''
                documentation = ''
                typ = 'untyped'
                samples = []
                allowed_names = []
                yield build_metric(sample[0], documentation, typ, [sample])
            else:
                samples.append(sample)

    if name != '':
        yield build_metric(name, documentation, typ, samples)



import sys
import time
import json
import argparse
import logging

from prov_p4.bmv2_controller import BMV2Controller


class ProvCollector:
    """This class collects the provenance from the data plane and sends it to the ProvManager for processing"""

    def __init__(
        self, reg_prov_details, cp_ip, thrift_port, switch_name, pkt_type, index_map
    ) -> None:
        self.reg_prov_details = reg_prov_details
        self.cp_ip = cp_ip
        self.dp_port = thrift_port
        self.controller = BMV2Controller(self.cp_ip, self.dp_port)
        self.agent_id = self.controller.get_switch_info()
        self.switch_name = switch_name
        self.pkt_type = pkt_type
        self.index_map = index_map
        self.last_count_index = {"packet_in": 0, "packet_out": 0}
        self.last_count_table_rule = 0
        for prov_reg_name in self.reg_prov_details:
            self.last_count_index[prov_reg_name] = 0
    
    def start(self): 
        POLL_TIME = 2  # In seconds
        try:
            while True:
                self._read_prov_regs()
                time.sleep(POLL_TIME)
        except KeyboardInterrupt:
            print("Closing Provenance Collector")

    def _read_prov_regs(self):
        for _reg_name in self.reg_prov_details:
            reg_name = _reg_name.replace(".")[1]
            if (
                reg_name == "packet_in"
                or (reg_name == "table_rule")
                or (reg_name == "packet_out")
            ):
                continue
            prov_reg_name = f"prov_{reg_name}"
            try:
                register_data = self.controller.read_register(prov_reg_name)
            except Exception as e:
                print(
                    f"[ERROR] Unable to read prov register (name - {prov_reg_name})...exiting"
                )
                logger.error(
                    f"Unable to read prov register (name - {prov_reg_name})...exiting",
                    e,
                )
                sys.exit(1)
            # Send Prov Agent event first
            record_agent(self.switch_name, self.agent_id, "Switch", time.time())
            prov_width_list = [
                ["index", self.reg_prov_details[reg_name]["index_size"]],
                ["value", self.reg_prov_details[reg_name]["value_size"]],
                ["activity", ACTIVITY_SIZE],
                ["version", VERSION_SIZE],
            ]
            # 

def main():
    """This script takes the provenance details, extracts the DP provenance, parses it
    and sends it to the RabbitMQ for processing using SPADE"""
    parser = argparse.ArgumentParser(
        description="This script requires the provenance detail file and thrift port as input."
    )
    parser.add_argument("--prov-details", type=str, help="Provenance details JSON")
    parser.add_argument("--switch-name", type=str, help="Switch Name (s0,s1...)")
    parser.add_argument(
        "--pkt-type", type=str, help="PacketIn/PacketOut field types JSON configuration"
    )
    parser.add_argument("--index-map", type=str, help="Index map JSON configuration")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-f", "--thrift-port-file", type=str, help="Path to the thrift port file."
    )
    group.add_argument("-p", "--thrift-port", type=int, help="Thrift port directly.")
    args = parser.parse_args()

    if args.thrift_port_file:
        with open(args.thrift_port_file, "r") as file:
            line = file.readline().strip()
            thrift_port = int(line)
    elif args.thrift_port is not None:
        thrift_port = args.thrift_port
    cp_ip = "localhost"
    args = parser.parse_args()
    # Read all the configuration files - provenance details, index map, and packet annotations

    with open(args.prov_details, "r") as file:
        reg_prov_details = json.load(file)
    with open(args.index_map, "r") as file:
        index_map = json.load(file)
    with open(args.pkt_type, "r") as file:
        pkt_type = json.load(file)
    print(reg_prov_details)
    print(index_map)
    print(pkt_type)
    collector = ProvCollector(
        reg_prov_details, cp_ip, thrift_port, args.switch_name, pkt_type, index_map
    )
    collector.start()


if __name__ == "__main__":
    main()



# This file contains the code to analyze and instrument the P4 program for provenance capture.
import json
import logging
from typing import Any, Set
from collections import defaultdict

from p4_action import P4Action
from p4_register_operation import P4RegisterOperation
# from provp4_augmenter import ProvP4Augmenter

fh_log = logging.FileHandler("provp4_analysis.log")
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s", "%Y-%m-%d")
fh_log.setFormatter(formatter)
logger = logging.getLogger("provp4-analysis")
logger.setLevel(logging.DEBUG)
logger.addHandler(fh_log)

class ProvP4ProgramAnalyzer:
    def __init__(self, compiled_p4: str, prog_name: str):
        self.compiled_p4 = compiled_p4
        self.prog_name = prog_name
        self.reg_bitwidth = {}
        self.fields_size_map = {}
        self.register_list = defaultdict[Any, dict](dict)
        self.augmentation_map = defaultdict[Any, dict](dict)
        self.actions_list = []
        self.json_prov_repr = {}

    def analyze_program(self):
        """
        Main runner function for the P4 program analysis.
        """
        self.extract_size_info()
        self.read_actions, self.write_actions = self.extract_stateful_info()
        self.extract_stateful_actions_info()
        self.write_prov_info_to_file()

    def extract_size_info(self):
        """
        Extract the size information from the P4 program.
        """
        # 1) Register bitwidths
        for register_defs in self.compiled_p4.get("register_arrays", []):
            self.reg_bitwidth[register_defs["name"]] = register_defs["bitwidth"]

        # 2) Header field sizes
        for header in self.compiled_p4.get("header_types", []):
            self.fields_size_map[header["name"]] = {}
            for field in header.get("fields", []):
                self.fields_size_map[header["name"]][field[0]] = field[1]
    def _parse_register_params(self, params, operation):
        if operation == "READ":
            value, name, index = params
        elif operation == "WRITE":
            name, index, value = params
        return name, index, value

    def extract_stateful_info(self):
        """
        Extract the stateful information from the P4 program.
        """
        total_actions, total_reg_writes, total_reg_reads = 0, 0, 0
        write_actions_set = set()
        read_actions_set = set()

        for action in self.compiled_p4["actions"]:
            total_actions += 1
            action_name = action.get("name")
            for primitive in action["primitives"]:
                if primitive["op"] == "register_write":
                    total_reg_writes += 1
                    write_actions_set.add(action_name)
                elif primitive["op"] == "register_read":
                    total_reg_reads += 1
                    read_actions_set.add(action_name)
                # Add support for clone operations in the future
                # elif primitive["op"] == "clone_ingress_pkt_to_egress":
                    # self.clone_details_map[primitive["source_info"]["filename"]].update(
                    #     {
                    #         primitive["source_info"]["line"]: [
                    #             action["id"],
                    #             action["name"],
                    #         ]
                    #     }
                    # )
        read_actions = sorted(read_actions_set)
        write_actions = sorted(write_actions_set)
        logger.info(f"Total Actions : {total_actions}")
        logger.info(f"Total Register writes : {total_reg_writes}")
        logger.info(f"Total Register reads : {total_reg_reads}")
        logger.info(f"Total actions w/ register writes: {len(write_actions)}")
        logger.info(f"Read actions: {read_actions}")
        logger.info(f"Write actions: {write_actions}")
        return read_actions, write_actions

    def extract_stateful_actions_info(self):
        """
        Extract the stateful actions information from the P4 program.
        """
        read_set: Set[str] = set(self.read_actions)
        write_set: Set[str] = set(self.write_actions)
        for action in self.compiled_p4.get("actions", []):
            action_name = action.get("name")
            action_id = action.get("id")
            if not action_name or action_id is None:
            # skip malformed entries
                continue
            needs_reads = action_name in read_set
            needs_writes = action_name in write_set

            current_action = P4Action(action_name, action_id)
            primitives = action.get("primitives", [])
            for prim_idx, primitive in enumerate(primitives):
                op = primitive.get("op")

                # Only process register_read and register_write operations
                if op != "register_read" and op != "register_write":
                    continue
                
                if op == "register_write" and not needs_writes:
                    continue
                if op == "register_read" and not needs_reads:
                    continue
                reg_op = self._build_register_operation(
                    action=action,
                    primitive=primitive,
                    prim_idx=prim_idx,
                    mode="WRITE" if op == "register_write" else "READ",
                )
                if reg_op is None:
                    continue
                # Attach to the correct list on the action object
                if reg_op.op_type == "WRITE":  # adjust attribute name if different
                    current_action.register_writes.append(reg_op)
                else:
                    # Ensure your P4Action has register_reads; if not, create it or use register_ops
                    current_action.register_reads.append(reg_op)
            # Only append if we actually added any operations
            if getattr(current_action, "register_writes", []) or getattr(current_action, "register_reads", []):
                self.actions_list.append(current_action)

    def _build_register_operation(
        self,
        *,
        action,
        primitive,
        prim_idx: int,
        mode: str,  # "READ" or "WRITE"
    ):
        action_name = action.get("name")
        action_id = action.get("id")
        src = primitive.get("source_info", {})

        line = src.get("line")
        filename = src.get("filename")
        fragment = src.get("source_fragment")
        if action_name is None or action_id is None or line is None:
            return None
        uid = int(f"{action_id}{line}") # Use a hash funtion
        name, index, value = self._parse_register_params(primitive.get("parameters", []), mode)
        reg_name = name.get("value")
        reg_name_key = reg_name.replace(".", "_") # P4 does not allow . in register names
        if not reg_name:
            return None
        # Keep track of which actions touch which registers
        self.register_list[reg_name_key].update({action_id: action_name})
        bitwidth = self.reg_bitwidth.get(reg_name)
        if bitwidth is None:
            raise Exception("Bitwidth not found in compiled JSON")
        current_reg_op = P4RegisterOperation(
            mode,
            reg_name_key,
            action_name,
            action_id,
            uid,
            index,
            value,
            fragment,
            line,
            bitwidth,
            filename,
        )
        # Compute sizes and update register_list
        current_reg_op.compute_prov_size(self.fields_size_map)
        if current_reg_op.reg_size_exceeded:
            # Recompute the sizes
            current_reg_op.compute_prov_size(self.fields_size_map)
        if current_reg_op.index_size is None or current_reg_op.value_size is None:
            raise Exception("Index or value size not found in compiled JSON")
        if current_reg_op.index_size == 0 or current_reg_op.value_size == 0:
            raise Exception("Index or value size is 0")
        self.register_list[reg_name_key].update(
            {"index_size": current_reg_op.index_size, "value_size": current_reg_op.value_size}
        )

        # Provenance definitions + JSON representation for augmentation
        current_reg_op.make_prov_definition()
        current_reg_op.prov_register_declaration()
        self.json_prov_repr.update(current_reg_op.get_json_repr())
        print(self.json_prov_repr)
        # Save augmentation info
        self._save_for_augmentation(
            filename,
            line,
            current_reg_op.uid,
            current_reg_op.log_write_str,
            current_reg_op.reg_decl,
            current_reg_op.prov_index_str,
        )
        # print(self.augmentation_map)
        return current_reg_op

    def _save_for_augmentation(
        self,
        filename: str,
        line_no: str,
        statement_id: int,
        write_str: str,
        reg_decl_str: str,
        prov_index_str: str,
    ):
        self.augmentation_map[filename].update(
            {line_no: [statement_id, write_str, reg_decl_str, prov_index_str]}
        )
    def write_prov_info_to_file(self):
        """
        Write the provenance information to a file.
        """
        # TODO - Add PacketIn and PacketOut information later
        with open(f"{self.prog_name}_prov_info.json", "w") as file:
            json.dump(self.register_list, file)

# if __name__ == "__main__":
    


"""networkx-query."""

from .definition import Evaluator, ParserException
from .parser import prepare_query
from .query import search_edges, search_nodes
from .relationship import PathCriteria, search_direct_relationships, search_relationships

__all__ = [
    "search_nodes",
    "search_edges",
    "search_direct_relationships",
    "prepare_query",
    "ParserException",
    "Evaluator",
    "search_relationships",
    "PathCriteria",
]



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

"""
This collection of classes helps collect authorization attributes about various resources
"""

from typing import Dict, Any, List
from collections import defaultdict

from fim.user.topology import ExperimentTopology
from fim.user.node import Node
from fim.user.network_service import NetworkService, ServiceType
from fim.user.component import Component
from fim.graph.slices.networkx_asm import NetworkxASM
from fim.slivers.topology_diff import TopologyDiff
from fim.slivers.network_service import NetworkServiceSliver
from fim.slivers.network_node import NodeSliver, NodeType
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.attached_components import ComponentSliver
from fim.view_only_dict import ViewOnlyDict


class LogCollector:
    """
    This class collects information for use in logging facility use. It can collect information
    from ExperimentTopology object, ASMs and TopologyDiffs. The dictionary returned by attributes
    property has the following structure:
    'nodes': list of VM Capacity objects
    'core_count': total cores
    'vm_count': total VMs
    'p4_count': total P4 switches
    'components': dictionary of ComponentType:quantity
    'services': list of tuples (ServiceType, bw)
    'facilities': set of facility port names (unique)
    'sites': set of site names (unique)
    """
    METHOD_LUT = {
        ExperimentTopology: "topo",
        NetworkxASM: "asm",
        TopologyDiff: "topology_diff",
        NodeSliver: "node_sliver",
        NetworkServiceSliver: "ns_sliver",
        ComponentSliver: "component_sliver",
        BaseSliver: "base_sliver",
        Node: "node",
        NetworkService: "ns",
        Component: "component"
    }

    def __init__(self):
        self._attributes = dict()
        self._attributes['nodes'] = list()
        self._attributes['core_count'] = 0
        self._attributes['vm_count'] = 0
        self._attributes['p4_count'] = 0
        self._attributes['components'] = dict()
        self._attributes['services'] = list()
        self._attributes['facilities'] = set()
        self._attributes['sites'] = set()

    @property
    def attributes(self):
        return ViewOnlyDict(self._attributes)

    def _collect_attributes_from_component_sliver(self, comp: ComponentSliver):
        cnt = self._attributes['components'].get(str(comp.resource_type), 0)
        self._attributes['components'][str(comp.resource_type)] = cnt + 1

    def _collect_attributes_from_node_sliver(self, sliver: NodeSliver):

        if sliver.resource_type == NodeType.VM:
            self._attributes['vm_count'] += 1
            cap = None
            if sliver.capacity_allocations:
                cap = sliver.capacity_allocations
            elif sliver.capacities:
                cap = sliver.capacities
            if cap:
                self._attributes['core_count'] += cap.core
                self._attributes['nodes'].append(cap)
        elif sliver.resource_type == NodeType.Switch:
            self._attributes['p4_count'] += 1
        elif sliver.resource_type == NodeType.Facility:
            self._attributes['facilities'].add(sliver.get_name())
        if sliver.site:
            self._attributes['sites'].add(sliver.site)
        if sliver.attached_components_info:
            for c in sliver.attached_components_info.list_devices():
                self._collect_attributes_from_component_sliver(c)

    def _collect_attributes_from_ns_sliver(self, sliver: NetworkServiceSliver):
        bw = sliver.capacities.bw if sliver.capacities else 0
        service_tup = (str(sliver.resource_type), bw)
        self._attributes['services'].append(service_tup)
        if sliver.site:
            self._attributes['sites'].add(sliver.site)

    def _collect_attributes_from_base_sliver(self, sliver: BaseSliver):
        if isinstance(sliver, NetworkServiceSliver):
            self._collect_attributes_from_ns_sliver(sliver)
        elif isinstance(sliver, NodeSliver):
            self._collect_attributes_from_node_sliver(sliver)

    def _collect_attributes_from_topo(self, topo: ExperimentTopology):
        for n in topo.nodes.values():
            self._collect_attributes_from_node(n)
        for ns in topo.network_services.values():
            self._collect_attributes_from_ns(ns)
        for fac in topo.facilities.values():
            self._attributes['facilities'].add(fac.name)

    def _collect_attributes_from_component(self, comp: Component):
        self._collect_attributes_from_component_sliver(comp.get_sliver())

    def _collect_attributes_from_node(self, node: Node):
        self._collect_attributes_from_node_sliver(node.get_sliver())

    def _collect_attributes_from_ns(self, ns: NetworkService):
        self._collect_attributes_from_ns_sliver(ns.get_sliver())

    def _collect_attributes_from_asm(self, asm: NetworkxASM):
        # convert to experiment topology
        asm.validate_graph()
        t = ExperimentTopology(graph_string=asm.serialize_graph())
        t.validate()
        self._collect_attributes_from_topo(t)

    def _collect_attributes_from_topology_diff(self, diff: TopologyDiff):
        """
        Use topology diff and only look at 'added' elements
        """
        for n in diff.added.nodes:
            self._collect_attributes_from_node(n)
        for ns in diff.added.services:
            self._collect_attributes_from_ns(ns)
        for c in diff.added.components:
            self._collect_attributes_from_component(c)

    def collect_resource_attributes(self, *, source: ExperimentTopology or
                                    NetworkxASM or
                                    TopologyDiff or
                                    NetworkService or Node or Component or
                                    NetworkServiceSliver or NodeSliver or ComponentSliver or None):
        """
        Take an ExperimentTopology or an ASM graph or TopologyDiff
        and extract all relevant attributes needed for logging, storing them
        as a Dict
        :param source: a source of attributes (ExperimentTopology, ASM or TopologyDiff)
        :return: dictionary
        """
        assert source

        method_suffix = LogCollector.METHOD_LUT.get(source.__class__)

        if not method_suffix:
            raise LogCollectorException(f'Unsupported resource type {source.__class__} '
                                        f'for collecting attributes')
        self.__getattribute__('_collect_attributes_from_' + method_suffix)(source)

    def __str__(self):
        """
        produce everything that goes after 'with' in a logging message
        """
        comp = ' compute ' + ','.join(['vms:' + str(self._attributes['vm_count']),
                                       'cores:' + str(self._attributes['core_count']),
                                       'p4s:' + str(self._attributes['p4_count'])])
        sites = facilities = components = services = vmdetails = ''

        if len(self._attributes['sites']):
            sites = ' sites ' + ','.join(self._attributes['sites'])

        if len(self._attributes['facilities']):
            facilities = ' facilities ' + ','.join(self._attributes['facilities'])

        if len(self._attributes['components']) > 0:
            cdict = self._attributes['components']
            clist = [k + ':' + str(v) for k, v in cdict.items()]
            components = ' components ' + ','.join(clist)

        if len(self._attributes['services']) > 0:
            slist = [t[0] + ':' + str(t[1]) for t in filter(lambda n: n[0] != 'OVS', self._attributes['services'])]
            services = ' services ' + ','.join(slist)

        # aggregate VM capacity information
        if len(self._attributes['nodes']) > 0:
            vmdict = defaultdict(int)
            for cap in self._attributes['nodes']:
                # these are capacity_allocation or capacity objects (if capacity_allocation is set it is preferred)
                vmdict[f'C{cap.core}/R{cap.ram}/D{cap.disk}'] += 1
            vmdetails = ' vmdetails ' + ','.join([i[0] + ':' + str(i[1]) for i in vmdict.items()])

        return ';'.join(filter(lambda n: len(n) > 0, [comp, sites, facilities, components, services, vmdetails]))

    def __repr__(self):
        return str(self.attributes)


class LogCollectorException(Exception):

    def __init__(self, msg: str):
        super().__init__(f"LogCollectorException: {msg}")


class P4Action:
    def __init__(self, name: str, uid: int) -> None:
        self.name = name
        self.uid = uid
        # List containing all the writes for a specific action
        self.register_writes = []
        self.register_reads = []
        self.prov_width = (
            0  # Total width of the prov record - size(index) + size(value)
        )

    def __str__(self) -> str:
        return f"P4Action Name: {self.name}"


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2021 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import Dict, Any, List, Tuple

import os
import json

from fim.view_only_dict import ViewOnlyDict

from .capacities_labels import Capacities


class InstanceCatalog:
    """
    This class helps list available VM instance sizes in FABRIC and
    map between instance size names and actual requirements for cores
    RAM and disk. The class relies on a resource file
    fim/slivers/data/instance_sizes.json
    """
    __catalog_instance = None

    def __init__(self):
        pass

    @staticmethod
    def __read_catalog() -> ViewOnlyDict:
        if InstanceCatalog.__catalog_instance is None:
            catalog_file = os.path.join(os.path.dirname(__file__), 'data', 'instance_sizes.json')
            with open(catalog_file) as f:
                catalog = json.load(f)
            assert isinstance(catalog, dict)
            InstanceCatalog.__catalog_instance = {k: Capacities(**v) for (k, v) in catalog.items()}
        return ViewOnlyDict(InstanceCatalog.__catalog_instance)

    def get_instance_capacities(self, *, instance_type: str) -> Capacities or None:
        """
        Return Capacities object for a given instance name or None
        :param instance_type:
        :return:
        """
        assert instance_type is not None
        c = self.__read_catalog()
        return c.get(instance_type, None)

    def map_capacities_to_instance(self, *, cap: Capacities) -> str:
        """
        Return closest instance (name) for these capacities. Return largest
        possible instance request exceeds what is possible.
        :param cap:
        :return: instance type name
        """
        assert cap is not None
        c = self.__read_catalog()
        keys = list(c.keys())
        values = list(c.values())
        candidates = list(filter(lambda x: (x.core >= cap.core and x.ram >= cap.ram and x.disk >= cap.disk),
                                 values))
        candidates.sort()
        if len(candidates) > 0:
            return keys[values.index(candidates[0])]
        # return the largest instance
        return keys[-1]

    def list_instances(self) -> ViewOnlyDict:
        """
        List available instance types and their capacities
        :return:
        """
        return self.__read_catalog()







#
# initialize the catalogs
from fim.slivers.component_catalog import ComponentCatalog
from fim.slivers.instance_catalog import InstanceCatalog

ComponentCatalog().populate_catalog_models_and_types()
InstanceCatalog()


#!/usr/bin/env python3
import enum
import getopt
import sys
from ipaddress import IPv4Address, IPv6Address
from datetime import datetime

from scapy.all import sniff
from scapy.layers.l2 import Ether
from scapy.layers.inet import IP
from scapy.layers.inet6 import IPv6, IPv6ExtHdrDestOpt
from scapy.packet import Packet, Raw, bind_layers
from scapy.fields import ShortField, LongField, XShortField, IntField

PROT_CMA = 254
CMA_ID = b"\x65\x87"

if __name__ == "__main__":

    class CMA(Packet):
        name = "CMA"
        fields_desc = [
            ShortField("len", 0),
            LongField("rx_time", 0),
            ShortField("mon_id", 0),
            ShortField("rx_port_id", 0),
            IntField("res", 0),
            LongField("uid_time", 0),
            ShortField("uid_mon", 0),
            ShortField("uid_port", 0),
            ShortField("time_chk", 0),
            XShortField("cma", CMA_ID)
        ]

    class PktLayers(enum.IntEnum):
        ETHER_O = 0
        IPv6_O = 1
        CMA = 2
        ETHER_I = 3
        IP_I = 4
        PROT_I = 5
        RAW_I = 6

    class Flow:
        def __init__(self, txport, rxport, ip_src, ip_dst, ip_prot, prot_sport, prot_dport):
            self.txport = txport
            self.rxport = rxport
            self.ip_src = ip_src
            self.ip_dst = ip_dst
            self.ip_prot = ip_prot
            self.prot_sport = prot_sport
            self.prot_dport = prot_dport

        def __hash__(self):
            return hash((self.txport, self.rxport, self.ip_src, self.ip_dst, self.ip_prot,
                        self.prot_sport, self.prot_dport))
        
        def __eq__(self, other):
            return (self.txport, self.rxport, self.ip_src, self.ip_dst, self.ip_prot,
                        self.prot_sport, self.prot_dport) == (other.txport, other.rxport, other.ip_src, other.ip_dst, other.ip_prot,
                                                              other.prot_sport, other.prot_dport)

    bind_layers(IPv6, CMA, nh=PROT_CMA)
    bind_layers(CMA, Ether)
    flowptr = 0
    flows = {}

    def analyze_packet(pkt: Packet, monitors: dict[int, dict[int, tuple[str, str]]]):
        global flowptr
        global flows
        size = None
        monitor_id = None
        monitor_iface_id = None
        uid = None
        eth_src = None
        eth_dst = None
        eth_type = None
        ip_ver = None
        ip_src = None
        ip_dst = None
        ip_prot = None
        prot_sport = None
        prot_dport = None
        time = None

        if CMA in pkt:
            print("Received packet")
            base = pkt['CMA']
            monitor_id = base.mon_id
            monitor_iface_id = base.rx_port_id
            uid = f'{base.uid_mon}-{base.uid_port}-{base.uid_time}'
            size = base.len - 16
            epoch = base.rx_time / 1e9
            time = datetime.fromtimestamp(epoch).strftime("%Y-%m-%d_%H:%M:%S.%f")

            h_eth = base['Ether']
            eth_type = hex(h_eth.type)
            h_ip = None
            if IP in base:
                h_ip = base['IP']
                ip_src = IPv4Address(h_ip.src)
                ip_dst = IPv4Address(h_ip.dst)
                ip_prot = h_ip.proto
            elif IPv6 in base:
                h_ip = base['IPv6']
                ip_src = IPv6Address(h_ip.src)
                ip_dst = IPv6Address(h_ip.dst)
                ip_prot = h_ip.nh
            if ip_prot == 6:
                h_tcp = h_ip['TCP']
                prot_sport = h_tcp.sport
                prot_dport = h_tcp.dport
            elif ip_prot == 17:
                h_tcp = h_ip['UDP']
                prot_sport = h_tcp.sport
                prot_dport = h_tcp.dport

            tx_id = None
            rx_id = None
            if monitor_iface_id == 1:
                tx_id = 1
                rx_id = 2
            else:
                tx_id = 2
                rx_id = 1
            tx_port = monitors[monitor_id][tx_id]
            rx_port = monitors[monitor_id][rx_id]
            thisflow = Flow(tx_port, rx_port, ip_src, ip_dst, ip_prot, prot_sport, prot_dport)
            flowhash = hash(thisflow)
            fid = None
            if flowhash in flows:
                fid = flows[flowhash]
                # print(f'Previous flow {fid}')
            else:
                fid = flowptr
                flowptr += 1
                flows[flowhash] = fid
                # print(f'New flow {fid}')
                with open("spade_pipe", 'a') as pipe:
                    print(f'type:Artifact id:{fid} eth.type:{eth_type} ip.src:{ip_src} ip.dst:{ip_dst} ip.prot:{ip_prot} prot.sport:{prot_sport} prot.dport:{prot_dport}\n')
                    pipe.write(f'type:Artifact id:{fid} eth.type:{eth_type} ip.src:{ip_src} ip.dst:{ip_dst} ip.prot:{ip_prot} prot.sport:{prot_sport} prot.dport:{prot_dport}\n')
            # print('Writing spade edges')
            with open("spade_pipe", 'a') as pipe:
                print(f'type:Used from:{rx_port} to:{fid} pkt_id:{uid} size:{size} time:{time} epoch:{epoch}\n')
                pipe.write(f'type:Used from:{rx_port} to:{fid} pkt_id:{uid} size:{size} time:{time} epoch:{epoch}\n')
                print(f'type:WasGeneratedBy from:{fid} to:{tx_port} pkt_id:{uid} size:{size} time:{time} epoch:{epoch}\n')
                pipe.write(f'type:WasGeneratedBy from:{fid} to:{tx_port} pkt_id:{uid} size:{size} time:{time} epoch:{epoch}\n')

    optpairs, args = getopt.getopt(sys.argv[1:], "m")
    monitors = {}
    ana_iface = ""
    with open("spade_pipe", 'a') as pipe:
        ana_iface = args[0]
        idx = 1
        id = -1
        while idx < len(args):
            iface_parts = args[idx].split('@', 1)
            if len(iface_parts) == 1:
                id = args[idx]
                monitors[int(id)] = {}
            else:
                iface_id = iface_parts[0]
                node = iface_parts[1].split('-', 1)[0]
                print(f'type:Agent id:{node}\n')
                pipe.write(f'type:Agent id:{node}\n')
                print(f'type:Process id:{iface_parts[1]}\n')
                pipe.write(f'type:Process id:{iface_parts[1]}\n')
                print(f'type:WasControlledBy from:{iface_parts[1]} to:{node}\n')
                pipe.write(f'type:WasControlledBy from:{iface_parts[1]} to:{node}\n')
                monitors[int(id)][int(iface_id)] = iface_parts[1]
            idx += 1
    print(monitors)
    print(ana_iface)
    sniff(prn=lambda x: analyze_packet(x, monitors), iface=ana_iface)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import Dict, Any, List, Tuple

import uuid

from fim.view_only_dict import ViewOnlyDict
from .model_element import ModelElement, ElementType, TopologyException
from .component import Component, ComponentType
from .network_service import NetworkService
from .interface import Interface

from ..graph.abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException

from ..slivers.network_node import NodeSliver
from ..slivers.network_node import NodeType
from ..slivers.network_service import ServiceType, NetworkServiceInfo
from ..slivers.component_catalog import ComponentModelType
from ..slivers.capacities_labels import CapacityHints, Location
from ..slivers.maintenance_mode import MaintenanceInfo
from ..slivers.interface_info import InterfaceType


class Node(ModelElement):
    """
    A basic node of any topology. In addition to public methods the following calls
    return various dictionaries or lists:
    node.components - returns a dictionary of components by name
    node.interfaces - returns a dictionary of interfaces by name
    node.interface_list - returns a list of interfaces
    node.direct_interfaces - returns a dictionary of direct interfaces, i.e. those
    not attached to components (mostly relevant to switches, not servers
    or VMs)
    network_services - returns a dictionary of network_services (mostly relevant
    to switches, not servers or VMs)
    """

    def __init__(self, *, name: str, node_id: str = None, topo: Any, etype: ElementType = ElementType.EXISTING,
                 ntype: NodeType = None, site: str = None, ns_info: NetworkServiceInfo = None,
                 check_existing: bool = False, **kwargs):
        """
        Don't call this method yourself, call topology.add_node()
        node_id will be generated if not provided for experiment topologies

        :param name:
        :param node_id:
        :param topo:
        :param etype: is this supposed to exist or new should be created
        :param ntype: node type if it is new
        :param site: node site
        :param ns_info: NetworkServiceInfo structure
        :param check_existing: for an existing node check if it is in the graph
        :param kwargs: any additional properties
        """
        assert name is not None
        assert topo is not None

        if etype == ElementType.NEW:
            # cant use isinstance as it would create circular import dependencies
            # node id myst be specified for new nodes in substrate topologies
            if str(topo.__class__) == "<class 'fim.user.topology.SubstrateTopology'>" and \
                    node_id is None:
                raise TopologyException("When adding new nodes to substrate topology nodes you must specify static Node ID")
            if node_id is None:
                node_id = str(uuid.uuid4())
            super().__init__(name=name, node_id=node_id, topo=topo)
            if ntype is None:
                raise TopologyException("When creating nodes you must specify NodeType")
            if site is None:
                raise TopologyException("When creating nodes you must specify site")

            sliver = NodeSliver()
            sliver.node_id = self.node_id
            sliver.set_name(self.name)
            sliver.set_type(ntype)
            sliver.set_site(site)
            sliver.network_service_info = ns_info
            sliver.set_properties(**kwargs)
            self.topo.graph_model.add_network_node_sliver(sliver=sliver)
        else:
            assert node_id is not None
            super().__init__(name=name, node_id=node_id, topo=topo)
            if check_existing:
                # check that this node exists
                existing_node_id = self.topo.graph_model.find_node_by_name(node_name=name,
                                                                           label=str(ABCPropertyGraph.CLASS_NetworkNode))
                if existing_node_id != node_id:
                    raise TopologyException(f'Node name {name} is not unique within the topology')

    @property
    def site(self):
        return self.get_property('site') if self.__dict__.get('topo', None) is not None else None

    @site.setter
    def site(self, value: str):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('site', value)

    @property
    def type(self):
        return self.get_property('type') if self.__dict__.get('topo', None) is not None else None

    @property
    def image_type(self):
        return self.get_property('image_type') if self.__dict__.get('topo', None) is not None else None

    @image_type.setter
    def image_type(self, value: str):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('image_type', value)

    @property
    def image_ref(self):
        return self.get_property('image_ref') if self.__dict__.get('topo', None) is not None else None

    @image_ref.setter
    def image_ref(self, value: str):
        if self.__dict__.get('topo', None) is not None:
            imtype = self.get_property('image_type')
            # image ref and type have fate-sharing - neither can be null to be written into a graph
            self.set_properties(image_ref=value, image_type=imtype)

    @property
    def management_ip(self):
        # no user-facing setter for this one
        return self.get_property('management_ip') if self.__dict__.get('topo', None) is not None else None

    @property
    def capacity_hints(self):
        return self.get_property('capacity_hints') if self.__dict__.get('topo', None) is not None else None

    @capacity_hints.setter
    def capacity_hints(self, value: CapacityHints):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('capacity_hints', value)

    @property
    def location(self):
        return self.get_property('location') if self.__dict__.get('topo', None) is not None else None

    @location.setter
    def location(self, value: Location):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('location', value)

    @property
    def maintenance_info(self):
        """
        Do not try to manipulate maintenance_info (PerNodeMaintenanceInfo) via this property,
        assign a complete object or unset
        """
        return self.get_property('maintenance_info') if self.__dict__.get('topo', None) is not None else None

    @maintenance_info.setter
    def maintenance_info(self, value: MaintenanceInfo):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('maintenance_info', value)

    @property
    def components(self):
        return self.__list_components()

    @property
    def interfaces(self):
        return self.__list_interfaces()

    @property
    def interface_list(self):
        return self.__list_of_interfaces()

    @property
    def direct_interfaces(self):
        return self.__list_direct_interfaces()

    @property
    def network_services(self):
        return self.__list_network_services()

    def validate_constraints(self):
        """
        Validate node constraints - properties
        """

        nstype = self.type

        # check properties
        req_props = NodeSliver.NodeConstraints[nstype].required_properties
        forb_props = NodeSliver.NodeConstraints[nstype].forbidden_properties
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        node_sliver = self.topo.graph_model.node_sliver_from_graph_properties_dict(node_properties)
        for rp in req_props:
            if not node_sliver.property_exists(rp) or \
                    (node_sliver.property_exists(rp) and not node_sliver.get_property(rp)):
                raise TopologyException(f"Node of type {nstype} must have property {rp} set")
        for fp in forb_props:
            if node_sliver.property_exists(fp) and node_sliver.get_property(fp):
                raise TopologyException(f"Node of type {nstype} must NOT have property {fp} set")

    def get_sliver(self) -> NodeSliver:
        """
        Get a deep sliver representation of this node from graph
        :return:
        """
        return self.topo.graph_model.build_deep_node_sliver(node_id=self.node_id)

    def get_property(self, pname: str) -> Any:
        """
        Retrieve a node property
        :param pname:
        :return:
        """
        assert pname is not None
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        node_sliver = self.topo.graph_model.node_sliver_from_graph_properties_dict(node_properties)
        return node_sliver.get_property(pname)

    def set_property(self, pname: str, pval: Any):
        """
        Set a node property or unset if pval is None
        :param pname:
        :param pval:
        :return:
        """
        if pval is None:
            self.unset_property(pname)
            return
        node_sliver = NodeSliver()
        node_sliver.set_property(prop_name=pname, prop_val=pval)
        # write into the graph
        prop_dict = self.topo.graph_model.node_sliver_to_graph_properties_dict(node_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    def set_properties(self, **kwargs):
        """
        Set multiple properties of the node
        :param kwargs:
        :return:
        """
        node_sliver = NodeSliver()
        node_sliver.set_properties(**kwargs)
        # write into the graph
        prop_dict = self.topo.graph_model.node_sliver_to_graph_properties_dict(node_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    @staticmethod
    def list_properties() -> Tuple[str]:
        return tuple(NodeSliver.list_properties())

    def add_component(self, *, name: str, node_id: str = None, ctype: ComponentType = None,
                      model: str = None, model_type: ComponentModelType = None,
                      network_service_node_id: str=None, interface_node_ids=None,
                      interface_labels=None, **kwargs) -> Component:
        """
        Add a component of specified type, model and name to this node. When working with substrate
        topologies you must specify the network_service_node_id and provide a list of interface node ids.
        :param name:
        :param node_id:
        :param ctype: ComponentType
        :param model: Model string (exact match)
        :param  model_type: ComponentModelType (combines ComponentType and Model)
        :param network_service_node_id: network service identifier for substrate models
        :param interface_node_ids: interface identifiers for substrate models
        :param interface_labels: list of labels for interfaces in substrate models
        :param kwargs: additional properties of the component
        :return:
        """
        assert name is not None
        # make sure name is unique within the node
        if name in self.__list_components().keys():
            raise TopologyException('Component names must be unique within node.')
        # add component node and populate properties
        c = Component(name=name, node_id=node_id, topo=self.topo, etype=ElementType.NEW,
                      ctype=ctype, model=model, comp_model=model_type,
                      network_service_node_id=network_service_node_id,
                      interface_node_ids=interface_node_ids, interface_labels=interface_labels,
                      parent_node_id=self.node_id, **kwargs)
        return c

    def add_storage(self, *, name: str, node_id: str = None, **kwargs) -> Component:
        """
        Add storage as a component (only in ASMs). Note that storage shows up as a component
        on the list of node components. Typically you want to specify labels=Labels(local_name='volume name')
        as part of keyargs so an existing volume can get attached.
        :param name:
        :param node_id:
        :param kwargs: additional properties of the component
        :return:
        """
        assert name is not None
        if str(self.topo.__class__) != "<class 'fim.user.topology.ExperimentTopology'>":
            raise TopologyException('Storage can be added as a component only in ExperimentTopologies')
        # make sure name is unique within the node
        if name in self.__list_components().keys():
            raise TopologyException('Component names must be unique within node.')
        # add component node and populate properties
        c = Component(name=name, node_id=node_id, topo=self.topo, etype=ElementType.NEW,
                      ctype=ComponentType.Storage, model=NodeType.NAS.name,
                      parent_node_id=self.node_id, **kwargs)
        return c

    def add_network_service(self, *, name: str, node_id: str = None, nstype: ServiceType, **kwargs) -> NetworkService:
        """
        Add a network service to node ( needed in substrate topologies)
        :param name:
        :param node_id:
        :param layer:
        :param kwargs:
        :return:
        """
        assert name is not None
        # make sure name is unique within the node
        if name in self.__list_network_services().keys():
            raise TopologyException('NetworkService names must be unique within node.')
        ns = NetworkService(name=name, node_id=node_id, parent_node_id=self.node_id, topo=self.topo,
                            etype=ElementType.NEW, nstype=nstype, **kwargs)
        return ns

    def remove_component(self, name: str) -> None:
        """
        Remove a component from the node (and network services and their interfaces)
        :param name:
        :return:
        """
        assert name is not None
        node_id = self.topo.graph_model.find_component_by_name(parent_node_id=self.node_id,
                                                               component_name=name)

        for i in self.components[name].interface_list:
            # disconnect if connected to a network service
            peers = i.get_peers(itype=InterfaceType.ServicePort)
            if peers:
                if len(peers) == 1:
                    # disconnect from its parent service
                    self.topo.get_parent_element(peers[0]).disconnect_interface(i)
                else:
                    raise TopologyException(f'Interface {i.name} has more than one peer, this is a model error.')
        # remove component, its network service and interfaces (if present)
        self.topo.graph_model.remove_component_with_nss_cps_and_links(node_id=node_id)

    def remove_network_service(self, name: str) -> None:
        """
        Remove a network service from the node (and all its interfaces)
        :param name:
        :return:
        """
        assert name is not None

        node_id = self.topo.graph_model.find_ns_by_name(parent_node_id=self.node_id,
                                                        nsname=name)
        self.topo.graph_model.remove_ns_with_cps_and_links(node_id=node_id)

    def remove_storage(self, name: str) -> None:
        """
        Removes attached storage component
        """
        self.remove_component(name)

    def __get_component_by_name(self, name: str) -> Component:
        """
        Find component of a node by its name, return Component object
        :param name:
        :return:
        """
        assert name is not None
        node_id = self.topo.graph_model.find_component_by_name(parent_node_id=self.node_id,
                                                               component_name=name)
        return Component(name=name, node_id=node_id, topo=self.topo)

    def _get_component_by_id(self, node_id: str) -> Component:
        """
        Get component of a node by its node_id, return Component object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return Component(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                         topo=self.topo)

    def __get_ns_by_name(self, name: str) -> NetworkService:
        """
        Find NetworkService of a node by its name, return NetworkService object
        :param name:
        :return:
        """
        assert name is not None
        node_id = self.topo.graph_model.find_ns_by_name(parent_node_id=self.node_id, component_name=name)
        return NetworkService(name=name, node_id=node_id, topo=self.topo)

    def __get_ns_by_id(self, node_id: str) -> NetworkService:
        """
        Get a network service of a node by its node_id, return NetworkService object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return NetworkService(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                              topo=self.topo)

    def __get_interface_by_id(self, node_id: str) -> Interface:
        """
        Get an interface of a node by its node_id, return Interface object
        :param node_id:
        :return:
        """
        assert node_id is not None
        clazzes, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        assert ABCPropertyGraph.CLASS_ConnectionPoint in clazzes
        return Interface(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                         topo=self.topo)

    def __list_components(self) -> ViewOnlyDict:
        """
        List all Components children of a node in the topology as a dictionary
        organized by component name. Modifying the dictionary will not affect
        the underlying model, but modifying Components in the dictionary will.
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_network_node_components(parent_node_id=self.node_id)
        ret = dict()
        for nid in node_id_list:
            c = self._get_component_by_id(nid)
            ret[c.name] = c
        return ViewOnlyDict(ret)

    def __list_network_services(self) -> ViewOnlyDict:
        """
        List all network service children of a node as a dictionary organized
        by network service name. Modifying the dictionary will not affect
        the underlying model, but modifying NetworkServices in the dictionary will.
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_network_node_or_component_nss(parent_node_id=self.node_id)
        ret = dict()
        for nid in node_id_list:
            c = self.__get_ns_by_id(nid)
            ret[c.name] = c
        return ViewOnlyDict(ret)

    def __list_direct_interfaces(self) -> ViewOnlyDict:
        """
        List all directly-attached interfaces of the node as a dictionary
        :return:
        """
        # immediately-attached interfaces
        node_if_list = self.topo.graph_model.get_all_node_or_component_connection_points(parent_node_id=self.node_id)
        ret = dict()
        for nid in node_if_list:
            i = self.__get_interface_by_id(nid)
            ret[i.name] = i
        return ViewOnlyDict(ret)

    def __list_interfaces(self) -> ViewOnlyDict:
        """
        List all interfaces of node and its components
        :return:
        """
        # immediately-attached interfaces
        # FIXME: note that because there could be a collision on interface name, this
        # may produce invalid results sometimes. Better to use list_of_interfaces
        node_if_list = self.topo.graph_model.get_all_node_or_component_connection_points(parent_node_id=self.node_id)
        direct_interfaces = dict()
        for nid in node_if_list:
            i = self.__get_interface_by_id(nid)
            direct_interfaces[i.name] = i
        cdict = self.__list_components()
        for k, v in cdict.items():
            comp_interfaces = v.interfaces
            direct_interfaces.update(comp_interfaces)
        return ViewOnlyDict(direct_interfaces)

    def __list_of_interfaces(self) -> Tuple[Interface]:
        """
        List all interfaces of node and its components
        :return:
        """
        node_if_list = self.topo.graph_model.get_all_node_or_component_connection_points(parent_node_id=self.node_id)
        direct_interfaces = list()
        for nid in node_if_list:
            i = self.__get_interface_by_id(nid)
            direct_interfaces.append(i)
        cdict = self.__list_components()
        for k, v in cdict.items():
            direct_interfaces.extend(v.interface_list)
        return tuple(direct_interfaces)

    def get_component(self, name: str):
        """
        Get a component by this name
        :param name:
        :return:
        """
        return self.__get_component_by_name(name)

    def __repr__(self):
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        node_sliver = self.topo.graph_model.node_sliver_from_graph_properties_dict(node_properties)
        return node_sliver.__repr__()

    def __str__(self):
        return self.__repr__()





"""Markdown filters with mistune

Used from markdown.py
"""
# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import base64
import mimetypes
import os
from html import escape
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Iterable, Match, Optional, Protocol, Tuple

import bs4
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexer import Lexer
from pygments.lexers import get_lexer_by_name
from pygments.util import ClassNotFound

from nbconvert.filters.strings import add_anchor

if TYPE_CHECKING:
    try:
        from mistune.plugins import Plugin
    except ImportError:

        class Plugin(Protocol):  # type: ignore[no-redef]
            """Mistune plugin interface."""

            def __call__(self, markdown: "Markdown") -> None:
                """Apply the plugin on the markdown document."""
                ...


try:  # for Mistune >= 3.0
    from mistune import (  # type:ignore[attr-defined]
        BlockParser,
        BlockState,
        HTMLRenderer,
        InlineParser,
        InlineState,
        Markdown,
        import_plugin,
    )

    MISTUNE_V3 = True
    MISTUNE_V3_ATX = "atx_heading" in BlockParser.SPECIFICATION

except ImportError:  # for Mistune >= 2.0
    import re

    from mistune import (  # type: ignore[attr-defined]
        PLUGINS,
        BlockParser,
        HTMLRenderer,
        InlineParser,
        Markdown,
    )

    MISTUNE_V3 = False
    MISTUNE_V3_ATX = False

    def import_plugin(name: str) -> "Plugin":  # type: ignore[misc]
        """Simple implementation of Mistune V3's import_plugin for V2."""
        return PLUGINS[name]  # type: ignore[no-any-return]


class InvalidNotebook(Exception):
    """An invalid notebook model."""


def _dotall(pattern: str) -> str:
    """Makes the '.' special character match any character inside the pattern, including a newline.

    This is implemented with the inline flag `(?s:...)` and is equivalent to using `re.DOTALL`.
    It is useful for LaTeX environments, where line breaks may be present.
    """
    return f"(?s:{pattern})"


if MISTUNE_V3:  # Parsers for Mistune >= 3.0.0

    class MathBlockParser(BlockParser):
        """This acts as a pass-through to the MathInlineParser. It is needed in
        order to avoid other block level rules splitting math sections apart.

        It works by matching each multiline math environment as a single paragraph,
        so that other rules don't think each section is its own paragraph. Inline
        is ignored here.
        """

        ATX_HEADING_WITHOUT_LEADING_SPACES = (
            r"^ {0,3}(?P<atx_1>#{1,6})(?!#+)(?P<atx_2>[ \t]*(.*?)?)$"
            if MISTUNE_V3_ATX
            else r"^ {0,3}(?P<axt_1>#{1,6})(?!#+)(?P<axt_2>[ \t]*(.*?)?)$"
        )

        MULTILINE_MATH = _dotall(
            # Display math mode, old TeX delimiter: $$ \sqrt{2} $$
            r"(?<!\\)[$]{2}.*?(?<!\\)[$]{2}"
            "|"
            # Display math mode, new LaTeX delimiter: \[ \sqrt{2} \]
            r"\\\\\[.*?\\\\\]"
            "|"
            # LaTeX environment: \begin{equation} \sqrt{2} \end{equation}
            r"\\begin\{(?P<math_env_name>[a-z]*\*?)\}.*?\\end\{(?P=math_env_name)\}"
        )

        SPECIFICATION = {
            **BlockParser.SPECIFICATION,
            (
                "atx_heading" if MISTUNE_V3_ATX else "axt_heading"
            ): ATX_HEADING_WITHOUT_LEADING_SPACES,
            "multiline_math": MULTILINE_MATH,
        }

        # Multiline math must be searched before other rules
        DEFAULT_RULES: ClassVar[Iterable[str]] = ("multiline_math", *BlockParser.DEFAULT_RULES)  # type: ignore[assignment]

        def parse_multiline_math(self, m: Match[str], state: BlockState) -> int:
            """Send mutiline math as a single paragraph to MathInlineParser."""
            matched_text = m[0]
            state.add_paragraph(matched_text)
            return m.end()

    class MathInlineParser(InlineParser):
        r"""This interprets the content of LaTeX style math objects.

        In particular this grabs ``$$...$$``, ``\\[...\\]``, ``\\(...\\)``, ``$...$``,
        and ``\begin{foo}...\end{foo}`` styles for declaring mathematics. It strips
        delimiters from all these varieties, and extracts the type of environment
        in the last case (``foo`` in this example).
        """

        # Display math mode, using older TeX delimiter: $$ \pi $$
        BLOCK_MATH_TEX = _dotall(r"(?<!\\)\$\$(?P<math_block_tex>.*?)(?<!\\)\$\$")
        # Display math mode, using newer LaTeX delimiter: \[ \pi \]
        BLOCK_MATH_LATEX = _dotall(r"(?<!\\)\\\\\[(?P<math_block_latex>.*?)(?<!\\)\\\\\]")
        # Inline math mode, using older TeX delimiter: $ \pi $  (cannot be empty!)
        INLINE_MATH_TEX = _dotall(r"(?<![$\\])\$(?P<math_inline_tex>.+?)(?<![$\\])\$")
        # Inline math mode, using newer LaTeX delimiter: \( \pi \)
        INLINE_MATH_LATEX = _dotall(r"(?<!\\)\\\\\((?P<math_inline_latex>.*?)(?<!\\)\\\\\)")
        # LaTeX math environment: \begin{equation} \pi \end{equation}
        LATEX_ENVIRONMENT = _dotall(
            r"\\begin\{(?P<math_env_name>[a-z]*\*?)\}"
            r"(?P<math_env_body>.*?)"
            r"\\end\{(?P=math_env_name)\}"
        )

        SPECIFICATION = {
            **InlineParser.SPECIFICATION,
            "block_math_tex": BLOCK_MATH_TEX,
            "block_math_latex": BLOCK_MATH_LATEX,
            "inline_math_tex": INLINE_MATH_TEX,
            "inline_math_latex": INLINE_MATH_LATEX,
            "latex_environment": LATEX_ENVIRONMENT,
        }

        # Block math must be matched first, and all math must come before text
        DEFAULT_RULES: ClassVar[Iterable[str]] = (
            "block_math_tex",
            "block_math_latex",
            "inline_math_tex",
            "inline_math_latex",
            "latex_environment",
            *InlineParser.DEFAULT_RULES,
        )  # type: ignore[assignment]

        def parse_block_math_tex(self, m: Match[str], state: InlineState) -> int:
            """Parse older TeX-style display math."""
            body = m.group("math_block_tex")
            state.append_token({"type": "block_math", "raw": body})
            return m.end()

        def parse_block_math_latex(self, m: Match[str], state: InlineState) -> int:
            """Parse newer LaTeX-style display math."""
            body = m.group("math_block_latex")
            state.append_token({"type": "block_math", "raw": body})
            return m.end()

        def parse_inline_math_tex(self, m: Match[str], state: InlineState) -> int:
            """Parse older TeX-style inline math."""
            body = m.group("math_inline_tex")
            state.append_token({"type": "inline_math", "raw": body})
            return m.end()

        def parse_inline_math_latex(self, m: Match[str], state: InlineState) -> int:
            """Parse newer LaTeX-style inline math."""
            body = m.group("math_inline_latex")
            state.append_token({"type": "inline_math", "raw": body})
            return m.end()

        def parse_latex_environment(self, m: Match[str], state: InlineState) -> int:
            """Parse a latex environment."""
            attrs = {"name": m.group("math_env_name"), "body": m.group("math_env_body")}
            state.append_token({"type": "latex_environment", "attrs": attrs})
            return m.end()

else:  # Parsers for Mistune >= 2.0.0 < 3.0.0

    class MathBlockParser(BlockParser):  # type: ignore[no-redef]
        """This acts as a pass-through to the MathInlineParser. It is needed in
        order to avoid other block level rules splitting math sections apart.
        """

        MULTILINE_MATH = re.compile(
            # Display math mode, old TeX delimiter: $$ \sqrt{2} $$
            r"(?<!\\)[$]{2}.*?(?<!\\)[$]{2}|"
            # Display math mode, new LaTeX delimiter: \[ \sqrt{2} \]
            r"\\\\\[.*?\\\\\]|"
            # LaTeX environment: \begin{equation} \sqrt{2} \end{equation}
            r"\\begin\{([a-z]*\*?)\}.*?\\end\{\1\}",
            re.DOTALL,
        )

        # Regex for header that doesn't require space after '#'
        AXT_HEADING = re.compile(r" {0,3}(#{1,6})(?!#+)(?: *\n+|([^\n]*?)(?:\n+|\s+?#+\s*\n+))")

        # Multiline math must be searched before other rules
        RULE_NAMES = ("multiline_math", *BlockParser.RULE_NAMES)  # type: ignore[attr-defined]

        def parse_multiline_math(self, m: Match[str], state: Any) -> Dict[str, str]:
            """Pass token through mutiline math."""
            return {"type": "multiline_math", "text": m.group(0)}

    class MathInlineParser(InlineParser):  # type: ignore[no-redef]
        r"""This interprets the content of LaTeX style math objects.

        In particular this grabs ``$$...$$``, ``\\[...\\]``, ``\\(...\\)``, ``$...$``,
        and ``\begin{foo}...\end{foo}`` styles for declaring mathematics. It strips
        delimiters from all these varieties, and extracts the type of environment
        in the last case (``foo`` in this example).
        """

        # Display math mode, using older TeX delimiter: $$ \pi $$
        BLOCK_MATH_TEX = _dotall(r"(?<!\\)\$\$(.*?)(?<!\\)\$\$")
        # Display math mode, using newer LaTeX delimiter: \[ \pi \]
        BLOCK_MATH_LATEX = _dotall(r"(?<!\\)\\\\\[(.*?)(?<!\\)\\\\\]")
        # Inline math mode, using older TeX delimiter: $ \pi $  (cannot be empty!)
        INLINE_MATH_TEX = _dotall(r"(?<![$\\])\$(.+?)(?<![$\\])\$")
        # Inline math mode, using newer LaTeX delimiter: \( \pi \)
        INLINE_MATH_LATEX = _dotall(r"(?<!\\)\\\\\((.*?)(?<!\\)\\\\\)")
        # LaTeX math environment: \begin{equation} \pi \end{equation}
        LATEX_ENVIRONMENT = _dotall(r"\\begin\{([a-z]*\*?)\}(.*?)\\end\{\1\}")

        RULE_NAMES = (
            "block_math_tex",
            "block_math_latex",
            "inline_math_tex",
            "inline_math_latex",
            "latex_environment",
            *InlineParser.RULE_NAMES,  # type: ignore[attr-defined]
        )

        def parse_block_math_tex(self, m: Match[str], state: Any) -> Tuple[str, str]:
            """Parse block text math."""
            # sometimes the Scanner keeps the final '$$', so we use the
            # full matched string and remove the math markers
            text = m.group(0)[2:-2]
            return "block_math", text

        def parse_block_math_latex(self, m: Match[str], state: Any) -> Tuple[str, str]:
            """Parse block latex math ."""
            text = m.group(1)
            return "block_math", text

        def parse_inline_math_tex(self, m: Match[str], state: Any) -> Tuple[str, str]:
            """Parse inline tex math."""
            text = m.group(1)
            return "inline_math", text

        def parse_inline_math_latex(self, m: Match[str], state: Any) -> Tuple[str, str]:
            """Parse inline latex math."""
            text = m.group(1)
            return "inline_math", text

        def parse_latex_environment(self, m: Match[str], state: Any) -> Tuple[str, str, str]:
            """Parse a latex environment."""
            name, text = m.group(1), m.group(2)
            return "latex_environment", name, text


class IPythonRenderer(HTMLRenderer):
    """An ipython html renderer."""

    def __init__(
        self,
        escape: bool = True,
        allow_harmful_protocols: bool = True,
        embed_images: bool = False,
        exclude_anchor_links: bool = False,
        anchor_link_text: str = "¶",
        path: str = "",
        attachments: Optional[Dict[str, Dict[str, str]]] = None,
        **lexer_options,
    ):
        """Initialize the renderer."""
        super().__init__(escape, allow_harmful_protocols)
        self.embed_images = embed_images
        self.exclude_anchor_links = exclude_anchor_links
        self.anchor_link_text = anchor_link_text
        self.path = path
        self.lexer_options = lexer_options
        if attachments is not None:
            self.attachments = attachments
        else:
            self.attachments = {}

    def block_code(self, code: str, info: Optional[str] = None) -> str:
        """Handle block code."""
        lang: Optional[str] = ""
        lexer: Optional[Lexer] = None

        if info:
            if info.startswith("mermaid"):
                return self.block_mermaidjs(code)

            try:
                if info.strip().split(None, 1):
                    lang = info.strip().split(maxsplit=1)[0]
                    lexer = get_lexer_by_name(lang, **self.lexer_options)
            except ClassNotFound:
                code = f"{lang}\n{code}"
                lang = None

        if not lang:
            return super().block_code(code, info=info)

        formatter = HtmlFormatter()
        return highlight(code, lexer, formatter)

    def block_mermaidjs(self, code: str) -> str:
        """Handle mermaid syntax."""
        return (
            """<div class="jp-Mermaid"><pre class="mermaid">\n"""
            f"""{code.strip()}"""
            """\n</pre></div>"""
        )

    def block_html(self, html: str) -> str:
        """Handle block html."""
        if self.embed_images:
            html = self._html_embed_images(html)

        return super().block_html(html)

    def inline_html(self, html: str) -> str:
        """Handle inline html."""
        if self.embed_images:
            html = self._html_embed_images(html)

        return super().inline_html(html)

    def heading(self, text: str, level: int, **attrs: Dict[str, Any]) -> str:
        """Handle a heading."""
        html = super().heading(text, level, **attrs)
        if self.exclude_anchor_links:
            return html
        return str(add_anchor(html, anchor_link_text=self.anchor_link_text))

    def escape_html(self, text: str) -> str:
        """Escape html content."""
        return escape(text, quote=False)

    def block_math(self, body: str) -> str:
        """Handle block math."""
        return f"$${self.escape_html(body)}$$"

    def multiline_math(self, text: str) -> str:
        """Handle mulitline math for older mistune versions."""
        return text

    def latex_environment(self, name: str, body: str) -> str:
        """Handle a latex environment."""
        name, body = self.escape_html(name), self.escape_html(body)
        return f"\\begin{{{name}}}{body}\\end{{{name}}}"

    def inline_math(self, body: str) -> str:
        """Handle inline math."""
        return f"${self.escape_html(body)}$"

    def image(self, text: str, url: str, title: Optional[str] = None) -> str:
        """Rendering a image with title and text.

        :param text: alt text of the image.
        :param url: source link of the image.
        :param title: title text of the image.

        :note: The parameters `text` and `url` are swapped in older versions
            of mistune.
        """
        if MISTUNE_V3:
            url = self._embed_image_or_attachment(url)
        else:  # for mistune v2, the first argument is the URL
            text = self._embed_image_or_attachment(text)

        return super().image(text, url, title)

    def _embed_image_or_attachment(self, src: str) -> str:
        """Embed an image or attachment, depending on the configuration.
        If neither is possible, returns the original URL.
        """

        attachment_prefix = "attachment:"
        if src.startswith(attachment_prefix):
            name = src[len(attachment_prefix) :]

            if name not in self.attachments:
                msg = f"missing attachment: {name}"
                raise InvalidNotebook(msg)

            attachment = self.attachments[name]
            # we choose vector over raster, and lossless over lossy
            preferred_mime_types = ("image/svg+xml", "image/png", "image/jpeg")
            for mime_type in preferred_mime_types:
                if mime_type in attachment:
                    return f"data:{mime_type};base64,{attachment[mime_type]}"
            # otherwise we choose the first mimetype we can find
            default_mime_type = next(iter(attachment.keys()))
            return f"data:{default_mime_type};base64,{attachment[default_mime_type]}"

        if self.embed_images:
            base64_url = self._src_to_base64(src)
            if base64_url is not None:
                return base64_url

        return src

    def _src_to_base64(self, src: str) -> Optional[str]:
        """Turn the source file into a base64 url.

        :param src: source link of the file.
        :return: the base64 url or None if the file was not found.
        """
        src_path = os.path.join(self.path, src)

        if not os.path.exists(src_path):
            return None

        with open(src_path, "rb") as fobj:
            mime_type, _ = mimetypes.guess_type(src_path)

            base64_data = base64.b64encode(fobj.read())
            base64_str = base64_data.replace(b"\n", b"").decode("ascii")

            return f"data:{mime_type};base64,{base64_str}"

    def _html_embed_images(self, html: str) -> str:
        parsed_html = bs4.BeautifulSoup(html, features="html.parser")
        imgs: bs4.ResultSet[bs4.Tag] = parsed_html.find_all("img")

        # Replace img tags's sources by base64 dataurls
        for img in imgs:
            src = img.attrs.get("src")
            if src is None:
                continue

            base64_url = self._src_to_base64(img.attrs["src"])
            if base64_url is not None:
                img.attrs["src"] = base64_url

        return str(parsed_html)


class MarkdownWithMath(Markdown):
    """Markdown text with math enabled."""

    DEFAULT_PLUGINS = (
        # "abbr",  (see https://github.com/jupyter/nbconvert/pull/1853)
        # "footnotes",
        "strikethrough",
        "table",
        "url",
        "task_lists",
        "def_list",
    )

    def __init__(
        self,
        renderer: HTMLRenderer,
        block: Optional[BlockParser] = None,
        inline: Optional[InlineParser] = None,
        plugins: Optional[Iterable["Plugin"]] = None,
    ):
        """Initialize the parser."""
        if block is None:
            block = MathBlockParser()
        if inline is None:
            if MISTUNE_V3:
                inline = MathInlineParser(hard_wrap=False)
            else:
                inline = MathInlineParser(renderer, hard_wrap=False)  # type: ignore[arg-type,misc]
        if plugins is None:
            plugins = (import_plugin(p) for p in self.DEFAULT_PLUGINS)

        super().__init__(renderer, block, inline, plugins)

    def render(self, source: str) -> str:
        """Render the HTML output for a Markdown source."""
        return str(super().__call__(source))


def markdown2html_mistune(source: str) -> str:
    """Convert a markdown string to HTML using mistune"""
    return MarkdownWithMath(renderer=IPythonRenderer(escape=False)).render(source)



from collections.abc import Iterable
from itertools import chain
from typing import Dict, List, NamedTuple, Optional, Tuple

from networkx import Graph

from .query import prepare_query, search_edges
from .utils import get_first_item, get_second_item

__all__ = ["search_direct_relationships", "PathCriteria", "search_relationships"]


def search_direct_relationships(
    graph: Graph, source: Optional[Dict] = None, edge: Optional[Dict] = None, target: Optional[Dict] = None
) -> Iterable[Tuple]:
    """Search direct relation ship.

    Arguments:
        graph (Graph): graph instance
        source (Optional[Dict]): optional source node query constraint
        edge (Optional[Dict]): optional edge query constraint
        target (Optional[Dict]): optional target node query constraint

    Returns:
        (Iterable[Tuple]): itrable tuple of edge

    """
    _iterable = search_edges(graph=graph, query=edge) if edge else graph.edges()

    if source:
        _predicate_source = prepare_query(source)
        _iterable = filter(lambda edge: _predicate_source(graph.nodes[edge[0]]), _iterable)

    if target:
        _predicate_target = prepare_query(target)
        _iterable = filter(lambda edge: _predicate_target(graph.nodes[edge[1]]), _iterable)

    return _iterable


class PathCriteria(NamedTuple):
    """Defines path criteria.

    Note:
        With `target={}, egde=...` you could wrote a criteria based on egde constraint only.

    Args:
        target (Dict): target node query constraint.
            `{}` means all nodes
        edge (Dict): optional edge query constraint
    """

    target: Dict
    edge: Optional[Dict] = None


def search_relationships(graph: Graph, source: Dict, *nodes: PathCriteria) -> Iterable[List]:
    """Search all valid path according specified template.

    Args:
        graph (Graph): graph instance
        source (Dict): source node query constraint
        nodes (PathCriteria): ordered list of path criteria

    Returns:
        (Iterable[Tuple]): all path which validate specification

    Raises:
        (RuntimeError): if a path criteria have no constraint
    """

    if not nodes:
        # cast single edge to path
        for a, b in search_direct_relationships(graph=graph, source=source):
            yield [a, b]
        return

    # Validate path template
    node_count = len(nodes)
    assert node_count >= 1

    # build pair of relations ships
    relationships = []

    for i in range(node_count):
        item = set(
            search_direct_relationships(
                graph=graph,
                source=source if i == 0 else nodes[i - 1].target,
                edge=nodes[i].edge,
                target=nodes[i].target,
            )
        )
        relationships.append(item)

        if not len(item):
            raise RuntimeError(f"No valid path from {i}th criteria.")

    # recursive
    def visit(path, node, i):
        for a, b in filter(lambda t: t[0] == node, relationships[i]):
            if i + 1 < node_count:
                yield from visit(path=path + [a], node=b, i=i + 1)
            else:
                yield path + [a, b]

    for a, b in relationships[0]:
        if node_count == 1:
            yield [a, b]
        else:
            yield from visit(path=[a], node=b, i=1)


def join_relationship(
    graph: Graph, source: Iterable[Tuple], target: Iterable[Tuple]
) -> Iterable[Tuple]:  # pragma: no cover
    """Join two relation ship.

    With  source = (a, b), target = (c, d)
    return (e, _) as e in source (e, b) and e in target( e, d)

    """

    _source = set(source)
    _target = set(target)

    _nodes = set(filter(get_first_item, _source)).intersection(set(filter(get_first_item, _target)))

    return chain(
        filter(lambda edge: get_first_item(edge) in _nodes, _source),
        filter(lambda edge: get_first_item(edge) in _nodes, _target),
    )


def chain_relationship(
    graph: Graph, source: Iterable[Tuple], target: Iterable[Tuple]
) -> Iterable[Tuple]:  # pragma: no cover
    """Chain two direct relation ship.

    return edge (_ e) or (e, _) as e in source(_, e) and e in target (e, _)

    """

    _source = set(source)
    _target = set(target)

    _nodes = set(filter(get_second_item, _source)).intersection(set(filter(get_first_item, _target)))

    return chain(
        filter(lambda edge: get_second_item(edge) in _nodes, _source),
        filter(lambda edge: get_first_item(edge) in _nodes, _target),
    )



#!/usr/bin/env python


from ..utils import floatToGoString
from ..validation import (
    _is_valid_legacy_labelname, _is_valid_legacy_metric_name,
)

CONTENT_TYPE_LATEST = 'application/openmetrics-text; version=1.0.0; charset=utf-8'
"""Content type of the latest OpenMetrics text format"""


def _is_valid_exemplar_metric(metric, sample):
    if metric.type == 'counter' and sample.name.endswith('_total'):
        return True
    if metric.type in ('gaugehistogram') and sample.name.endswith('_bucket'):
        return True
    if metric.type in ('histogram') and sample.name.endswith('_bucket') or sample.name == metric.name:
        return True
    return False


def generate_latest(registry):
    '''Returns the metrics from the registry in latest text format as a string.'''
    output = []
    for metric in registry.collect():
        try:
            mname = metric.name
            output.append('# HELP {} {}\n'.format(
                escape_metric_name(mname), _escape(metric.documentation)))
            output.append(f'# TYPE {escape_metric_name(mname)} {metric.type}\n')
            if metric.unit:
                output.append(f'# UNIT {escape_metric_name(mname)} {metric.unit}\n')
            for s in metric.samples:
                if not _is_valid_legacy_metric_name(s.name):
                    labelstr = escape_metric_name(s.name)
                    if s.labels:
                        labelstr += ', '
                else:
                    labelstr = ''
                
                if s.labels:
                    items = sorted(s.labels.items())
                    labelstr += ','.join(
                        ['{}="{}"'.format(
                            escape_label_name(k), _escape(v))
                            for k, v in items])
                if labelstr:
                    labelstr = "{" + labelstr + "}"
                    
                if s.exemplar:
                    if not _is_valid_exemplar_metric(metric, s):
                        raise ValueError(f"Metric {metric.name} has exemplars, but is not a histogram bucket or counter")
                    labels = '{{{0}}}'.format(','.join(
                        ['{}="{}"'.format(
                            k, v.replace('\\', r'\\').replace('\n', r'\n').replace('"', r'\"'))
                            for k, v in sorted(s.exemplar.labels.items())]))
                    if s.exemplar.timestamp is not None:
                        exemplarstr = ' # {} {} {}'.format(
                            labels,
                            floatToGoString(s.exemplar.value),
                            s.exemplar.timestamp,
                        )
                    else:
                        exemplarstr = ' # {} {}'.format(
                            labels,
                            floatToGoString(s.exemplar.value),
                        )
                else:
                    exemplarstr = ''
                timestamp = ''
                if s.timestamp is not None:
                    timestamp = f' {s.timestamp}'
                if _is_valid_legacy_metric_name(s.name):
                    output.append('{}{} {}{}{}\n'.format(
                        s.name,
                        labelstr,
                        floatToGoString(s.value),
                        timestamp,
                        exemplarstr,
                    ))
                else:
                    output.append('{} {}{}{}\n'.format(
                        labelstr,
                        floatToGoString(s.value),
                        timestamp,
                        exemplarstr,
                    ))
        except Exception as exception:
            exception.args = (exception.args or ('',)) + (metric,)
            raise

    output.append('# EOF\n')
    return ''.join(output).encode('utf-8')


def escape_metric_name(s: str) -> str:
    """Escapes the metric name and puts it in quotes iff the name does not
    conform to the legacy Prometheus character set.
    """
    if _is_valid_legacy_metric_name(s):
        return s
    return '"{}"'.format(_escape(s))


def escape_label_name(s: str) -> str:
    """Escapes the label name and puts it in quotes iff the name does not
    conform to the legacy Prometheus character set.
    """
    if _is_valid_legacy_labelname(s):
        return s
    return '"{}"'.format(_escape(s))


def _escape(s: str) -> str:
    """Performs backslash escaping on backslash, newline, and double-quote characters."""
    return s.replace('\\', r'\\').replace('\n', r'\n').replace('"', r'\"')



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import List, Tuple, Any, Dict
import enum

import json


class PathRepresentationType(enum.Enum):
    """
    Path Representation type - can be represented as a list or as a graph
    """
    Path = enum.auto()   # PathInfo is a pair of uni-directional lists
    Graph = enum.auto()  # PathInfo is a graph reference (graph ID)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class Path:
    """
    Path object expressed as two lists of entities on the path
    """
    def __init__(self):
        self.a2z = None
        self.z2a = None

    def set(self, *, a2z: List = None, z2a: List = None):
        self.a2z = a2z
        self.z2a = z2a

    def set_symmetric(self, a2z: List):
        assert a2z is not None
        assert isinstance(a2z, list)
        self.a2z = a2z
        self.z2a = a2z.copy()
        self.z2a.reverse()

    def get(self) -> Tuple[List, List]:
        """
        Return a tuple a2z, z2a
        :return:
        """
        return self.a2z, self.z2a

    def to_dict(self) -> Dict:
        return {'a2z': self.a2z, 'z2a': self.z2a}

    @classmethod
    def from_dict(cls, d: Dict):
        assert isinstance(d, dict)
        assert 'a2z' in d.keys() and 'z2a' in d.keys()
        pi = cls()
        pi.set(a2z=d['a2z'], z2a=d['z2a'])
        return pi

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return json.dumps(self.to_dict())


class PathInfo:
    """
    PathInfo encodes information about paths. Most of the time lists (a2z, z2a) are used,
    however we can also do it as an external graph (by reference).
    """
    def __init__(self, ptype: PathRepresentationType = PathRepresentationType.Path):
        self.payload = None
        self.type = ptype

    def set(self, payload) -> None:
        """
        Path payload can be a Path or id of a graph built externally (string name is payload)
        :param payload:
        :return:
        """
        if self.type == PathRepresentationType.Path:
            assert(isinstance(payload, Path))
        else:
            assert(isinstance(payload, str))
        self.payload = payload

    def get(self) -> Tuple[PathRepresentationType, Any]:
        """
        Return a tuple of type, payload of Path (tuple of lists or a graph id)
        :return:
        """
        return self.type, self.payload

    @staticmethod
    def type_from_str(etype: str) -> PathRepresentationType or None:
        if etype is None:
            return None
        for t in PathRepresentationType:
            if etype == str(t):
                return t

    def to_json(self) -> str:
        """
        Convert to JSON dictionary:
        { "type": "list"|"graph", "payload": <payload> }
        payload can be a graph identifier string or a dict with 'a2z' and 'z2a' lists
        :return:
        """
        json_dict = dict()
        json_dict['type'] = str(self.type)
        json_dict['payload'] = self.payload if self.type == PathRepresentationType.Graph else self.payload.to_dict()
        return json.dumps(json_dict)

    @classmethod
    def from_json(cls, json_string: str):
        """
        Set fields from json string and returns a new object
        :param json_string:
        :return: object
        """
        if json_string is None or len(json_string) == 0:
            return None
        d = json.loads(json_string)
        ptype_str = d.get('type', None)
        if ptype_str is None:
            return None

        ptype = PathInfo.type_from_str(ptype_str)
        ret = cls(ptype)
        ret.payload = d['payload'] if ptype == PathRepresentationType.Graph else Path.from_dict(d['payload'])
        return ret

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.to_json()


class ERO(PathInfo):
    """
    Explicit Route Object class which adds 'strict' flag compared to PathInfo
    """
    def __init__(self, etype: PathRepresentationType = PathRepresentationType.Path, strict: bool = False):
        # default is 'loose' list
        super().__init__(etype)
        self.strict = strict

    def get_strict(self) -> bool:
        """
        Get the ERO strict flag
        :return:
        """
        return self.strict

    def to_json(self) -> str:
        """
        Convert to JSON dictionary:
        { "type": "Path"|"Graph", "payload": <payload> }
        payload can be a graph identifier string or a dict with 'a2z' and 'z2a' lists
        :return:
        """
        json_dict = dict()
        json_dict['type'] = str(self.type)
        json_dict['strict'] = str(self.strict)
        json_dict['payload'] = self.payload if self.type == PathRepresentationType.Graph else self.payload.to_dict()
        return json.dumps(json_dict)

    @classmethod
    def from_json(cls, json_string: str):
        """
        Set fields from json string and returns a new object
        :param json_string:
        :return: object
        """
        if json_string is None or len(json_string) == 0:
            return None
        d = json.loads(json_string)
        ptype_str = d.get('type', None)
        if ptype_str is None:
            return None

        ptype = PathInfo.type_from_str(ptype_str)
        ret = cls(ptype)
        ret.payload = d['payload'] if ptype == PathRepresentationType.Graph else Path.from_dict(d['payload'])
        ret.strict = True if d.get('strict', None) in {'True', 'true'} else False
        return ret

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.to_json()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import Any, List, Tuple

import uuid

from .model_element import ModelElement, ElementType, TopologyException

from ..slivers.interface_info import InterfaceType, InterfaceSliver
from ..graph.abc_property_graph import ABCPropertyGraph
from ..slivers.capacities_labels import Labels
from ..view_only_dict import ViewOnlyDict


class Interface(ModelElement):

    def __init__(self, *, name: str, node_id: str = None, topo: Any,
                 etype: ElementType = ElementType.EXISTING,
                 parent_node_id: str = None,
                 itype: InterfaceType = None,
                 check_existing: bool = False,
                 **kwargs):
        """
        Don't call this method yourself, call node.add_interface()
        node_id will be generated if not provided for experiment topologies

        :param name:
        :param node_id:
        :param topo:
        :param etype: is this supposed to exist or new should be created
        :param parent_node_id: parent network service or parent interface
        :param itype: node type if it is new
        :param check_existing: check if the Interface exists in the graph
        :param kwargs: any additional properties
        """
        assert name is not None
        assert topo is not None

        if etype == ElementType.NEW:
            # cant use isinstance as it would create circular import dependencies
            # node id myst be specified for new nodes in substrate topologies
            if str(topo.__class__) == "<class 'fim.user.topology.SubstrateTopology'>" and \
                    node_id is None:
                raise TopologyException("When adding new nodes to substrate topology nodes you must specify static Node ID")
            if node_id is None:
                node_id = str(uuid.uuid4())
            if itype is None:
                raise TopologyException("When creating interfaces you must specify InterfaceType")
            super().__init__(name=name, node_id=node_id, topo=topo)
            sliver = InterfaceSliver()
            sliver.node_id = self.node_id
            sliver.set_name(self.name)
            sliver.set_type(itype)
            sliver.set_properties(**kwargs)

            self.topo.graph_model.add_interface_sliver(parent_node_id=parent_node_id, interface=sliver)
            self._interfaces = list()
        else:
            assert node_id is not None
            super().__init__(name=name, node_id=node_id, topo=topo)
            if check_existing and not self.topo.graph_model.check_node_name(node_id=node_id, name=name,
                                                                            label=ABCPropertyGraph.CLASS_ConnectionPoint):
                raise TopologyException(f"Interface with this id {node_id} and name {name} doesn't exist")
            # collect a list of interface nodes it attaches to for DedicatedPorts only
            self._interfaces = list()
            if self.type == InterfaceType.DedicatedPort:
                interface_list = self.topo.graph_model.get_all_child_connection_points(interface_id=self.node_id)
                name_id_tuples = list()
                # need to look up their names - a bit inefficient, need to think about this /ib
                for iff in interface_list:
                    _, props = self.topo.graph_model.get_node_properties(node_id=iff)
                    name_id_tuples.append((props[ABCPropertyGraph.PROP_NAME], iff))
                self._interfaces = [Interface(node_id=tup[1], topo=topo, name=tup[0]) for tup in name_id_tuples]

    @property
    def type(self):
        return self.get_property('type') if self.__dict__.get('topo', None) is not None else None

    def add_child_interface(self, *, name: str, node_id: str = None, **kwargs):
        """
        Add an interface to network service
        :param name:
        :param node_id:
        :param itype: interface type e.g. TrunkPort, AccessPort or VINT
        :param kwargs: additional parameters
        :return:
        """
        assert name is not None
        assert self.type is InterfaceType.DedicatedPort

        # check uniqueness
        all_names = [n.name for n in self._interfaces]
        if name in all_names:
            raise TopologyException(f'Sub Interface {name} is not unique within the interface')

        labels = kwargs.get('labels')
        if not labels or not labels.vlan:
            raise TopologyException(f'Vlan must be specified for Sub Interface within the interface')

        all_vlans = []
        for i in self._interfaces:
            if i.labels and i.labels.vlan:
                all_vlans.append(i.labels.vlan)

        if labels.vlan in all_vlans:
            raise TopologyException(f'Vlan in use by another Sub Interface within the interface')

        if self.labels:
            labels.local_name = self.labels.local_name
        else:
            raise TopologyException(f'Local name for the dedicated interface not available')

        iff = Interface(name=name, node_id=node_id, parent_node_id=self.node_id,
                        etype=ElementType.NEW, topo=self.topo, itype=InterfaceType.SubInterface,
                        **kwargs)

        self._interfaces.append(iff)
        return iff

    def remove_child_interface(self, *, name: str) -> None:
        """
        Remove an ServicePort interface from the network service.
        :param name:
        :return:
        """
        assert name is not None
        assert self.type is InterfaceType.DedicatedPort

        # cant use isinstance as it would create circular import dependencies
        #if str(self.topo.__class__) == "<class 'fim.user.topology.ExperimentTopology'>":
        #    raise TopologyException("Cannot remove child interface interface from Interface in Experiment topology")
        node_id = self.topo.graph_model.find_child_connection_point_by_name(parent_node_id=self.node_id,
                                                                            iname=name)

        self.topo.graph_model.remove_cp_and_links(node_id=node_id, delete_parent=False)

    def __list_interfaces(self) -> ViewOnlyDict:
        """
        List all interfaces of the network service as a dictionary
        :return:
        """
        ret = dict()
        for intf in self._interfaces:
            ret[intf.name] = intf
        return ViewOnlyDict(ret)

    def __list_of_interfaces(self) -> tuple:
        """
        Return a list of names of interfaces of network service
        :return:
        """
        return tuple(self._interfaces)

    @property
    def interface_list(self):
        """
        List of names of service interfaces
        """
        return self.__list_of_interfaces()

    @property
    def interfaces(self):
        """
        Dictionary name->Interface for all interfaces
        """
        return self.__list_interfaces()

    @property
    def peer_labels(self):
        return self.get_property('peer_labels') if self.__dict__.get('topo', None) is not None else None

    @peer_labels.setter
    def peer_labels(self, value: Labels):
        if self.__dict__.get('topo', None) is not None:
            self.set_property('peer_labels', value)

    def get_property(self, pname: str) -> Any:
        """
        Retrieve a interface property
        :param pname:
        :return:
        """
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        if_sliver = self.topo.graph_model.interface_sliver_from_graph_properties_dict(node_properties)
        return if_sliver.get_property(pname)

    def set_property(self, pname: str, pval: Any):
        """
        Set a interface property or unset if pval is None
        :param pname:
        :param pval:
        :return:
        """
        if pval is None:
            self.unset_property(pname)
            return
        if_sliver = InterfaceSliver()
        if_sliver.set_property(prop_name=pname, prop_val=pval)
        # write into the graph
        prop_dict = self.topo.graph_model.interface_sliver_to_graph_properties_dict(if_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    def set_properties(self, **kwargs):
        """
        Set multiple properties of the interface
        :param kwargs:
        :return:
        """
        if_sliver = InterfaceSliver()
        if_sliver.set_properties(**kwargs)
        # write into the graph
        prop_dict = self.topo.graph_model.interface_sliver_to_graph_properties_dict(if_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    @staticmethod
    def list_properties() -> Tuple[str]:
        return InterfaceSliver.list_properties()

    def get_peers(self, itype: InterfaceType = None) -> List[Any] or None:
        """
        Find 'peer' interfaces connected across a Link. Returns a list of Interface objects
        (matching optional itype if specified, otherwise all).
        :parm itype: optional type of peer interface we are looking for
        :return: peer Interface object or None
        """
        peer_ids = self.topo.graph_model.find_peer_connection_points(node_id=self.node_id)
        if peer_ids is None:
            return None
        ret = list()
        for peer_id in peer_ids:
            clazzes, node_props = self.topo.graph_model.get_node_properties(node_id=peer_id)
            assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
            assert ABCPropertyGraph.CLASS_ConnectionPoint in clazzes
            i = Interface(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=peer_id,
                          topo=self.topo)
            if itype is not None:
                if i.type == itype:
                    ret.append(i)
            else:
                ret.append(i)
        return ret

    def get_sliver(self) -> InterfaceSliver:
        """
        Get a deep sliver representation of this interface from graph
        :return:
        """
        return self.topo.graph_model.build_deep_interface_sliver(node_id=self.node_id)

    def __repr__(self):
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        if_sliver = self.topo.graph_model.interface_sliver_from_graph_properties_dict(node_properties)
        return if_sliver.__repr__()

    def __str__(self):
        return self.__repr__()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
#



def _jupyter_labextension_paths():
    return [{"src": "../labextension", "dest": "jupyter-leaflet"}]


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "../nbextension",
            "dest": "jupyter-leaflet",
            "require": "jupyter-leaflet/extension",
        }
    ]
